#include <SpeedyStepper.h>
#include <Wire.h>
#include "Adafruit_TCS34725.h"

#define NUMCOLS 30      // number of colours in 'library' of known colours
#define NUMSAMPLES 5    // number of samples to take in quick succession when brick first arrives
#define HOLDOFFMS 200L  // time guaranteed to wait after brick has passed sensor before stepper can move again

  // normalised to brightness (scaled to 0-255):
  // 1red, 2green, 3blue, 4white, 5black, 6purple, 7orange, 8yellow, 9grey, 10magenta, 
  // 11cyan, 12lt_green, 13lt_pink, 14pink, 15greypink, 16dk_grey, 17dk_blue, 18lt_blue, 19beige, 20teal,
  // 21sage, 22racing_green, 23v_lt_blue, 24lilac, 25chestnut, 26chocolate, 27mushroom, 28tan, 29ketchup
const uint8_t rlist[NUMCOLS] = {0, 245,  68,  42, 147, 127, 106, 233, 189, 138, 219,  72, 151, 167, 198, 205, 134,  72,  89, 170,  57, 125,  98, 126, 141, 185, 206, 171, 204, 233};
const uint8_t glist[NUMCOLS] = {0,  51, 221, 122, 161, 170, 106,  89, 157, 160,  66, 171, 193, 128,  89,  92, 162, 145, 152, 160, 189, 185, 196, 170, 122, 162, 121, 154, 133, 74};
const uint8_t blist[NUMCOLS] = {0,  51, 110, 223, 134, 170, 205,  53,  65, 146, 117, 176,  74, 143, 135, 120, 148, 200, 186, 106, 167, 129, 137, 143, 174, 115,  97, 108,  86, 74};
// output bin number for each colour: 0 white/grey, 1 red/orange, 2 greens, 3 blues, 4 purple/pink, 5 yellow, 6 browns, 7 black
const uint8_t bins[NUMCOLS]  = { 7, 1, 2, 3, 0, 0, 4, 1, 5, 0, 4, 3, 2, 4, 4, 4, 0, 3, 3, 6, 3, 2, 2, 0, 4, 6, 6, 6, 6, 1};
uint16_t      c_thresh = 65535;

unsigned long holdoffstarttime;

enum stateenum{ready, measure, wait, holdoff};  
volatile enum stateenum state;

#define STEPSPERREV     200       // steps per whole rotation
#define MICROSTEPS      2         // microstepping mode
#define MAXSPEED        2900      // steps/sec
#define ACCEL           31000     // steps/sec-squared

#define POSNS           8         // no. of diverter positions
#define POSNS_OVER_TWO  (POSNS / 2)
#define STEPSPERPOSN    (STEPSPERREV * MICROSTEPS / POSNS)

const int MOTOR_STEP_PIN = 3;       // pin D3 (stepper motor step pin)
const int MOTOR_DIRECTION_PIN = 4;  // pin D4 (stepper motor direction)
const int BRICK_DETECT_PIN = 2;     // pin D2 (h/w interrupt from light sensor)
const int FEEDER_MOTOR_PWM_PIN = 9; // pin D9 (PWM for vibrating feeder motor)

bool brickArrivedFlag = false;

// Initialise stepper driver
SpeedyStepper stepper;

/* Initialise colour sensor with specific int time and gain values */
Adafruit_TCS34725 tcs = Adafruit_TCS34725(TCS34725_INTEGRATIONTIME_2_4MS, TCS34725_GAIN_16X);

void setup() {
  pinMode(FEEDER_MOTOR_PWM_PIN, OUTPUT);

  // start serial port
  Serial.begin(57600);

  // setup stepper driver
  stepper.connectToPins(MOTOR_STEP_PIN, MOTOR_DIRECTION_PIN);
  stepper.setSpeedInStepsPerSecond(MAXSPEED);
  stepper.setAccelerationInStepsPerSecondPerSecond(ACCEL);
  // exercise the stepper at start, just to check it's all working ok.
  for (uint8_t i=0;i<2;i++){
    move_to_pos(0);
    delay(100);
    move_to_pos(4);
    delay(100);
  }
  move_to_pos(0);

  // start vibrating feeder motor
  setupTimer1();
  setPWM1A(0.60);
  delay(100);
  setPWM1A(0.40);

  // start colour sensor
  if (tcs.begin()) {
    Serial.println("Found sensor");
  } else {
    Serial.println("No TCS34725 found ... check your connections");
    while (1)
      ;
  }
  // measure brightness at start as an indication of 'no brick' threshold - might vary with ambient light levels, etc.
  c_thresh = tcs.read16(TCS34725_CDATAL);

  // setup interrupts
  attachInterrupt(digitalPinToInterrupt(BRICK_DETECT_PIN), ISR_brick_detect, CHANGE);
  
  // init state-machine
  state = ready;
}


// Configure Timer 1 (pins 9,10) to output 15kHz PWM
// Note: the built-in PWM frequencies are all either audibly noisy or too fast for the L298N driver.
void setupTimer1(){
    //Set PWM frequency to about 15kHz on pins 9,10 (timer 1 mode 10, no prescale, count to 533)
    TCCR1A = (1 << COM1A1) | (1 << COM1B1) | (1 << WGM11);
    TCCR1B = (1 << CS10) | (1 << WGM13);
    ICR1 = 533;
    OCR1A = 0;
    OCR1B = 0;
}

// Equivalent of analogWrite on pin 9, except f is in range 0.0 to 1.0
void setPWM1A(float f){
    f=f<0?0:f>1?1:f;
    OCR1A = (uint16_t)(533*f);
}

// Equivalent of analogWrite on pin 10, except f is in range 0.0 to 1.0
void setPWM1B(float f){
    f=f<0?0:f>1?1:f;
    OCR1B = (uint16_t)(533*f);
}

void loop() {
  uint16_t r, g, b, c;
  uint16_t rmax, gmax, bmax, cmax = 0;
  

  if (state == measure) {
    state = wait;

    // grab N samples back-to-back really quickly and choose the one corresponding to the brightest reflection, generally the best estimate of colour
    // For N==4, this method takes around 12-14msec.
    for (uint8_t i = 0; i<NUMSAMPLES; i++) {

      c = tcs.read16(TCS34725_CDATAL);
      r = tcs.read16(TCS34725_RDATAL);
      g = tcs.read16(TCS34725_GDATAL);
      b = tcs.read16(TCS34725_BDATAL);

      if (c > cmax) {
        cmax = c;
        rmax = r; gmax = g; bmax = b;
      }
    }

    // process it and move the diverter to the corresponding position
    processColourData(rmax, gmax, bmax, cmax);

    // get ready to go again
    cmax = 0;
  }

  // gets called when holdoff timers reaches zero
  if ((state == holdoff) & (millis()-holdoffstarttime > HOLDOFFMS)){
    state = ready;
  }

}



void ISR_brick_detect(void) {
  // gets called when a Lego piece passes between the colour sensor's LED and the photodiode

  if (digitalRead(BRICK_DETECT_PIN) == HIGH) {
    // rising-edge transition: brick arrives at sensor
    if (state == ready) {
      state = measure;
    }
  }
  else {
    // falling-edge transition: brick passes and leaves sensor

    // (re)start hold-off timer
    holdoffstarttime = millis();

    // update state-machine
    if (state == wait) {
      state = holdoff;
    }
  }
}


// Move diverter to new position
void move_to_pos(uint8_t new_pos) {
  static uint8_t curr_pos = 0;

  if (new_pos == curr_pos) {
    return;
  }

  int8_t move = new_pos - curr_pos;

  if (move > POSNS_OVER_TWO) {
    move = move - POSNS;
  } else if (move < -POSNS_OVER_TWO) {
    move = move + POSNS;
  }

  stepper.moveRelativeInSteps(move * STEPSPERPOSN);
  curr_pos = new_pos;
}


uint8_t getColourIndex(uint32_t r, uint32_t g, uint32_t b, uint32_t c) {
  
  // normalise r,g,b values to 0-255
  uint32_t mag = sqrt((uint32_t)r*(uint32_t)r + (uint32_t)g*(uint32_t)g + (uint32_t)b*(uint32_t)b);
  uint8_t rn = (255 * r) / mag;
  uint8_t gn = (255 * g) / mag;
  uint8_t bn = (255 * b) / mag;
  
  // Serial.print(rn, DEC); Serial.print(",");
  // Serial.print(gn, DEC); Serial.print(",");
  // Serial.print(bn, DEC); Serial.print(",");
  // Serial.println(c, DEC);  
  
  return findNearestColourIndex(rn, gn, bn, c);
}

uint8_t findNearestColourIndex(uint8_t rn, uint8_t gn, uint8_t bn, uint16_t c) {
  // return value 1 to (NUMCOLS-1) or zero for no piece detected.
  uint32_t dist, mindist = UINT32_MAX;
  uint8_t retval = 0;

  if (c >= (c_thresh+2)) {
    for (int8_t i = 1; i < NUMCOLS; i++) {
      int32_t dr = (int32_t)rn - (int32_t)rlist[i];
      int32_t dg = (int32_t)gn - (int32_t)glist[i];
      int32_t db = (int32_t)bn - (int32_t)blist[i];

      // 'redmean' algorithm for colour difference calculation
      int32_t r_mean = ((int32_t)rn + (int32_t)rlist[i]) / 2;
      dist = (2 + r_mean/256)*dr*dr + 4*dg*dg + (2 + (255-r_mean)/256)*db*db; // + 6*dc*dc;

      // // linear distance (squared)
      // dist = dr*dr + dg*dg + db*db;

    if ((i == 1) || (dist < mindist)) {
        mindist = dist;
        retval = i;
      }
    }
  }
  return retval;
}


void processColourData(uint16_t r, uint16_t g, uint16_t b, uint16_t c) {
  // take latest raw colour measurements and convert to index into colour LUT, or -1 for 'no Lego'
  uint8_t i_col = getColourIndex(r, g, b, c);

  if ((bins[i_col]==0) & (c < c_thresh+20)) {
    // if classified as white/grey/black and brightness is very low, put it in the black bin, #7.
      move_to_pos(7);
  } else {
      // otherwise proceed as normal.
      move_to_pos(bins[i_col]);
  }
}


